/* EMACS_MODES: c,fill=0 */
#include	"hoc.h"
#include	"y.tab.h"
#include	<stdio.h>

#define	pop2(p1,p2)	p2 = pop(); p1 = pop();

#define	NSTACK	256
static	Datum	stack[NSTACK];		/* the stack */
static	Datum	*stackp;		/* next free spot on stack */

#define NPROG	2000
Inst	prog[NPROG];			/* the machine */
Inst	*progp;				/* next free spot in machine */
Inst	*pc;				/* program counter */
Inst	*progbase = prog;		/* start of current subprogram */
int	returning;			/* switch: 1=> return stmt seen */

typedef	struct	Frame {			/* proc/func call stack frame */
	Symbol	*sp;			/* symbol table entry */
	Inst	*retpc;			/* where to resume after return */
	Datum	*argn;			/* n-th argument on stack */
	int	nargs;			/* number of arguments */
} Frame;

#define	NFRAME	100
Frame	frame[NFRAME];			/* frame stack */
Frame	*fp;				/* next free loc in frame stack */

/*
 *				**** INITCODE ****
 *
 * This routine initializes the stack and machine to empty
 *
 */

initcode()
{
	stackp = stack;			/* empty stack */
	progp = progbase;		/* after definitions */
	fp = frame;			/* empty frame stack */
	returning = 0;			/* not returning (yet) */
}

/*
 * the following two routines manipulate the stack
 *
 */

/*
 *				**** PUSH ****
 *
 * This routine pushes an item onto the stack
 *
 */

push(d)
Datum d;
{
	if (stackp >= &stack[NSTACK])
		execerror("stack overflow", (char *) 0);
	*stackp++ = d;
}

/*
 *				**** POP ****
 *
 * This routine pops and returns the next element from the stack
 *
 */

Datum
pop()
{
	if (stackp <= stack)
		execerror("stack underflow", (char *) 0);
	return *--stackp;
}

/*
 *				**** CODE ****
 *
 * This routine adds a new instruction or operand into the machine
 * the argument passed is either the pointer to a function which will
 *	execute the machine instruction,
 * or an argument, which is a pointer to an element on the stack
 *
 */

Inst *
code(f)
Inst f;
{
	Inst *oprogp = progp;

	if (progp >= &prog[NPROG])
		execerror("program too big", (char *) 0);
	*progp++ = f;
	return oprogp;
}

/*
 *				**** EXECUTE ****
 *
 * The routine that actually runs this machine.
 * The program counter (pc) should be pointing to the address of a function
 *	that will execute the next instruction of the machine.
 * This function is resonsible for incrementing pc so that it points to
 *	the next instruction.
 *
 * If this is not done, then mass confusion breaks forth.
 *
 */

execute(p)
Inst *p;
{
	for (pc = p; *pc != STOP && !returning; ) (*(*pc++))();
}

/*
 * The next two functions push values onto the stack
 *
 */

/*
 *				**** CONSTPUSH ****
 * This routine pushes a constant onto the stack
 * pc should be pointing to the next free location in the stack
 *
 */

constpush()		/* push constant onto stack */
{
	Datum d;

	d.val = ((Symbol *) *pc++)->u.val; /* value to Datum */
	push(d);			/* and placed on stack */
}

/*
 *				**** VARPUSH ****
 *
 * This routine pushes the variable that was parsed onto the stack.
 *
 */

varpush()		/* push variable onto stack */
{
	Datum d;			/* for stack item */

	d.sym = (Symbol *)(*pc++);	/* symbol to Datum */
	push(d);			/* and place on stack */
}

/*
 * The following routines implement the arithmetic operations of he machine
 * This routine is called when its address appears on the top of the stack.
 * Following it should be the operands
 * The result should be pushed back onto the stack
 *
 */

/*
 *				**** ADD ****
 * add the top two elements on the stack
 *
 */

add()
{
	Datum d1, d2;

	d2 = pop();			/* get second operand */
	d1 = pop();			/* get first operand */
	d1.val += d2.val;		/* add them together */
	push(d1);			/* and push it back on the stack */
}

/*
 *				**** SUB ****
 *
 * This routine pops and subtracts the top two values on the stack,
 * and places the result back on the stack
 *
 */

sub()
{
	Datum d1, d2;

	d2 = pop();
	d1 = pop();
	d1.val -= d2.val;
	push(d1);
}

/*
 *				**** MUL ****
 *
 * This routine pops and multiplies the top two values on the stack.
 * The result is pushed onto the stack.
 *
 */

mul()
{
	Datum d1, d2;

	d2 = pop();
	d1 = pop();
	d1.val *= d2.val;
	push(d1);
}

/*
 *				**** DIV ****
 *
 * This routine pops and divides the top two values on the stack.
 * The result is pushed onto the stack
 *
 */

div()
{
	Datum d1, d2;

	d2 = pop();
	d1 = pop();
	if (d2.val == 0.0) execerror("division by zero", (char *) 0);
	/* execerror does not return */
	d1.val /= d2.val;
	push(d1);
}

/*
 *				**** POWER ****
 *
 * This routine pops the top two elements on the stack
 *	and raises the first to the second power
 * the result is pushed on the stack.
 *
 */

power()
{
	Datum d1, d2;
	extern double Pow();

	d2 = pop();
	d1 = pop();
	d1.val = Pow(d1.val,d2.val);
	push(d1);
}

/*
 *				**** NEGATE ****
 *
 * This routine negates the top value on the stack.
 *
 */

negate()
{
	Datum d;

	d = pop();
	d.val = - d.val;
	push(d);
}

/*
 *				**** EVAL ****
 *
 * This routine evaluates a variable by looking up its value.
 * The address of the symbol table entry is on the top of the stack.
 * This address is popped, and the current value of the symbol replaces it
 *
 */

eval()
{
	Datum d;

	d = pop();			/* get symbol table entry */
	/* the following should never occur, but if it does,
	 * execerror will not return
	 */
	if (d.sym->type == UNDEF)	/* not defined yet */
		execerror("undefined variable", d.sym->name);
	d.val = d.sym->u.val;		/* get current value */
	push(d);			/* and push it onto the stack */
}

/*
 *				**** ASSIGN ****
 *
 * This routine performs an assignment.
 * The symbol table entry and new value should be next on the stack
 * When done, the value should be left on the stack
 *
 */

assign()
{
	Datum d1, d2;

	d1 = pop();			/* the variable */
	d2 = pop();			/* the value */
	/* the following should never happen,
	 * but if it does, execerror does not return
	 */
	if (d1.sym->type != VAR && d1.sym->type != UNDEF) /* error */
		execerror("assignment to non-variable", d1.sym->name);
	d1.sym->u.val = d2.val;		/* assign the value */
	d1.sym->type = VAR;		/* in case it was undefined */
	push(d2);			/* push value back on the stack */
}

/*
 *				**** PRINT ****
 *
 * This routine will pop the top value on the stack, then discard it
 *
 */

print()
{
	Datum d;

	d = pop();
	printf("\t%.8g\n", d.val);
}

/*
 *				**** BLTIN ****
 *
 * This routine evaluates a built in function on the top of the stack
 * What else is on the stack, and what happens to the stack is up to
 *	the called routine
 * The called routine is to return the result, which is pushed on the stack
 *
 */

bltin()
{
	Datum d;

	d = pop();			/* address of built in */
	/* the next statement says that *pc should be cast to
	 *	"pointer to a function returning a double", and
	 *	that resulting function executed with d.val as an argument
	 * (Fortunately, the above quote followed this function in the text)
	 */
	d.val = (*(double (*)())(*pc++))(d.val);
	push(d);
}

/*
 * The following routines perform logical comparisons with the top two
 *	elements on the stack.
 *	These values are popped, and the result is pushed on the stack
 *
 */

/*
 *				**** GT ****
 */

gt()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val > d2.val);
	push(d1);
}

/*
 *				**** LT ****
 */

lt()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val < d2.val);
	push(d1);
}

/*
 *				**** EQ ****
 */

eq()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val == d2.val);
	push(d1);
}

/*
 *				**** GE ****
 */

ge()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val >= d2.val);
	push(d1);
}

/*
 *				**** LE ****
 */

le()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val <= d2.val);
	push(d1);
}

/*
 *				**** NE ****
 */

ne()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val != d2.val);
	push(d1);
}

/*
 *				**** AND ****
 */

and()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val && d2.val);
	push(d1);
}

/*
 *				**** OR ****
 */

or()
{
	Datum d1, d2;

	pop2(d1,d2);
	d1.val = (double)(d1.val || d2.val);
	push(d1);
}

/*
 *				**** NOT ****
 */

not()
{
	Datum d;

	d = pop();			/* get value from stack */
	d.val = (double) (! d.val);
	push(d);
}

/*
 *				**** PREXPR ****
 *
 * This routine prints an expression, and is called when the print command
 *	is encountered
 */

prexpr()
{
	Datum d;

	d = pop();
	printf("%.8g ", d.val);
}

/*
 *				**** PRSTR ****
 *
 * This routine prints a string
 *
 */
prstr()
{
	printf ("%s", (char *) *pc++);
}

/*
 *				**** VARREAD ****
 *
 * This routine will read a value into a variable
 * It returns a zero on EOF, and a 1 otherwise
 *
 */

varread()
{
	Datum	d;
	extern	FILE	*fin;		/* input file number */
	Symbol	*var = (Symbol *) *pc++;
	int	keep_going = 1;

	/* Since I refuse to use goto's, this is a variation of the text */

	while (keep_going) {
		keep_going = 0;		/* normal case */
		switch (fscanf(fin,"%lf", &var->u.val)) {
		case EOF:		/* end of this file, try next */
			/* if no more input, return zero, else next file */
			if (!moreinput()) d.val = var -> u.val = 0.0;
			else keep_going = 1; /* try again */
			continue;
		case 0:			/* invalid number */
			execerror("non-number read into",var->name);
			continue;
		default:		/* valid number */
			d.val = 1.0;	/* valid return */
			continue;
		}
	}
	var->type = VAR;		/* pushing a variable on stack */
	push(d);			/* push it now */
}

/*
 *				**** WHILECODE ****
 *
 * This routine produces the code necessary to perform a while loop.
 *
 * A while loop is implemented as follows within the stack:
 *
 *	Loc	Contents
 *	-1	address of this routine (whilecode)
 *	0	loc on stack of loop body (X)
 *	1	loc on stack of next stmt (Y)
 *	2	code for conditional stmt
 *	  ...
 *	X-1	STOP instruction
 *	X	body of while loop
 *	  ...
 *	Y-1	STOP instruction
 *	Y	next stmt after while loop
 *
 * This routine will execute the conditional statement, and then test it.
 * If the result was true, it will execute the body of the loop.
 * It will continue to do so until the conditional evaluates to false.
 *
 * NOTE:
 *
 * When this routine is entered, pc points to the location in the stack
 *	AFTER the address of this routine.
 *
 */

whilecode()
{
	Datum d;
	Inst *savepc = pc;		/* loop body */

	/*
	 * This routine is not coded the same way as in the text.
	 * I did this because I tend to think this way instead of
	 * the way in the book. Both are logically identical, but
	 * my method takes less instructions.
	 */

	while (!returning) {			/* in case in proc/func */
		execute(savepc+2);	/* execute conditional */
		d = pop();		/* get result of conditional */
		if (!d.val) break;	/* if false, we are done */
		execute(*((Inst **)(savepc))); /* execute body of loop */
	}
	if (!returning)	pc = *((Inst **)(savepc+1));	/* update pc */
}

/*
 *				**** IFCODE ****
 *
 * This routine executes an if statement.
 *
 * An if statement is set up as follows on the stack:
 *
 *	Loc	Contents
 *	-1	address of this routine (ifcode)
 *	0	address of then part (X)
 *	1	address of else part (Y)
 *	2	address of next statement (Z)
 *	3	the conditional statement
 *	  ...
 *	X-1	STOP instruction
 *	X	the then part
 *	  ...
 *	Y-1	STOP instruction
 *	Y	else part
 *	  ...
 *	Z-1	STOP instruction
 *	Z	the next statement
 *
 * This routine will first execute the conditional statement.
 * If the result is true, it will execute the 'then' part,
 *	otherwise it will execute the 'else' part.
 * In either case, it will then execute the next statement.
 *
 * NOTE:
 *
 * When called, the stack pointer (pc) points to the location in the stack
 *	AFTER the address of this routine
 *
 * If no 'else' statement was supplied, then the address of the routine
 *	will contain a NULL
 *
 */

ifcode()
{
	Datum d;
	Inst *savepc = pc;		/* then part */

	execute(savepc+3);		/* execute the conditional ... */
	d = pop();			/* ... and get the result */
	if (d.val)			/* if conditional was true */
		execute(*((Inst **)(savepc))); /* execute the then part */
	else if (*((Inst **)(savepc+1))) /* addr of else part or NULL */
		execute(*((Inst **)(savepc+1))); /* else part supplied */
	if (!returning)	pc = *((Inst **)(savepc+2));	/* update pc */
}

/*
 *				**** DEFINE ****
 *
 * This routine adds a function and/or procedure into the symbol table
 *
 */

define(sp)
Symbol *sp;
{
	sp->u.defn = (Inst) progbase;	/* start of code */
	progbase = progp;		/* next code starts here */
}

/*
 *				**** CALL ****
 *
 * This routine performs a call to a function
 * Procedure calling is implemented in the stack as follows:
 *
 *	Loc	Description
 *	-1	CALL instruction
 *	0	pointer to Symbol table entry for procedure
 *	+1	number of arguments
 *
 * This routine will set up the frame entry on the next frame stack.
 * The frame entry is formatted as follows:
 *
 *	Name	Description
 *	sp	Pointer to the symbol table entry for the procedure
 *	retpc	Return address: pointer to the stack of where to return
 *	argn	Address of the N-th argument on the stack (see nargs)
 *	nargs	Number of arguments for procedure
 *
 */

call()
{
	Symbol *sp = (Symbol *)pc[0]; /* symbol tbl entry for proc */

	if (fp++ >= &frame[NFRAME-1])
		execerror(sp->name, "call nested too deeply");
	fp->sp = sp;			/* pointer to symbol table */
	fp->nargs = (int)pc[1];		/* number of arguments */
	fp->retpc = pc + 2;		/* the return address */
	fp->argn = stackp - 1;		/* addr of last argument */
	execute(sp->u.defn);		/* execute the func/proc */
	returning = 0;			/* reset 'returning' flag */
}

/*
 *				**** RET ****
 *
 * This routine implements the returning instruction from a procedure
 * It pops the arguments off the stack, sets the return address,
 *	pops the frame entry from the frame stack, and finally
 *	sets the returning flag
 *
 */

ret()
{
	int	i;

	for (i = 0; i < fp->nargs; i++) pop(); /* pop arguments */
	pc = (Inst *)fp->retpc;		/* set pc to return address */
	--fp;				/* pop frame entry */
	returning = 1;			/* set flag */
}

/*
 *				**** FUNCRET ****
 *
 * This routine executes a return from a function
 * I essentially performs a normal return, preserving the return value
 *
 */

funcret()
{
	Datum	d;

	if (fp->sp->type == PROCEDURE)
		execerror(fp->sp->name,"(proc) returns a value");
	d = pop();			/* save return value */
	ret();				/* perform a return */
	push(d);			/* and push the return value back */
}

/*
 *				**** PROCRET ****
 *
 * This routine performs a procedure return,
 *	which a return without any value
 *
 */

procret()
{
	if (fp->sp->type == FUNCTION)
		execerror(fp->sp->name," (func) returns no value");
	ret();				/* just return */
}

/*
 *				**** GETARG ****
 *
 * This routine will return a pointer to the arguments for the current proc
 * When called, pc points to the entry containing the number of arguments
 *
 */

double *
getarg()
{
	int nargs = (int) *pc++;	/* the number of arguments */

	if (nargs > fp->nargs) execerror(fp->sp->name,"not enought args");
	return &fp->argn[nargs - fp->nargs].val;
}

/*
 *				**** ARG ****
 *
 * This routine pushes an argument onto the stack
 *
 */

arg()
{
	Datum	d;

	d.val = *getarg();		/* argument value */
	push(d);			/* put in on the stack */
}

/*
 *				**** ARGASSIGN ****
 *
 * This routine stores the stop of the stack into an argument
 *
 */

argassign()
{
	Datum d;

	d = pop();			/* get top value from stack */
	push(d);			/* and put it back */
	*getarg() = d.val;		/* assign value to argument */
}
